"""Module grouping tests for the interpretaties search module."""
import pandas as pd
from owslib.fes import PropertyIsEqualTo
from pandas import DataFrame

from pydov.search.interpretaties import GecodeerdeLithologieSearch
from pydov.types.interpretaties import GecodeerdeLithologie
from tests.abstract import AbstractTestSearch

location_md_metadata = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/' \
    'md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/' \
    'fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/' \
    'wfsdescribefeaturetype.xml'
location_wfs_getfeature = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/' \
    'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/feature.xml'
location_dov_xml = \
    'tests/data/types/interpretaties/gecodeerde_lithologie' \
    '/gecodeerde_lithologie.xml'
location_xsd_base = \
    'tests/data/types/interpretaties/gecodeerde_lithologie/xsd_*.xml'


class TestGecodeerdeLithologieSearch(AbstractTestSearch):

    search_instance = GecodeerdeLithologieSearch()
    datatype_class = GecodeerdeLithologie

    valid_query_single = PropertyIsEqualTo(propertyname='Proefnummer',
                                           literal='kb22d56w-B165')

    inexistent_field = 'onbestaand'
    wfs_field = 'Proefnummer'
    xml_field = 'grondsoort'

    valid_returnfields = ('pkey_interpretatie',
                          'betrouwbaarheid_interpretatie')
    valid_returnfields_subtype = (
        'pkey_interpretatie', 'diepte_laag_van', 'diepte_laag_tot')
    valid_returnfields_extra = ('pkey_interpretatie', 'gemeente')

    df_default_columns = ['pkey_interpretatie', 'pkey_boring',
                          'betrouwbaarheid_interpretatie', 'x', 'y',
                          'start_interpretatie_mtaw',
                          'diepte_laag_van', 'diepte_laag_tot',
                          'hoofdnaam1_grondsoort', 'hoofdnaam2_grondsoort',
                          'bijmenging1_plaatselijk', 'bijmenging1_hoeveelheid',
                          'bijmenging1_grondsoort', 'bijmenging2_plaatselijk',
                          'bijmenging2_hoeveelheid', 'bijmenging2_grondsoort',
                          'bijmenging3_plaatselijk', 'bijmenging3_hoeveelheid',
                          'bijmenging3_grondsoort']

    def test_search_nan(self, mp_wfs, mp_get_schema,
                        mp_remote_describefeaturetype, mp_remote_md,
                        mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        self.search_instance.search(
            query=self.valid_query_single)

    def test_search_customreturnfields(self, mp_get_schema,
                                       mp_remote_describefeaturetype,
                                       mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with custom return fields.

        Test whether the output dataframe is correct.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType .
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'pkey_boring'))

        assert isinstance(df, DataFrame)

        assert list(df) == ['pkey_interpretatie', 'pkey_boring']

        assert not pd.isnull(df.pkey_boring[0])

    def test_search_xml_resolve(self, mp_get_schema,
                                mp_remote_describefeaturetype,
                                mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'diepte_laag_tot'))

        assert df.diepte_laag_tot[0] == 4.

    def test_search_multiple_return(self, mp_get_schema,
                                    mp_remote_describefeaturetype,
                                    mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method returning multiple (sub)elements of the same subject.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie',
                           'hoofdnaam1_grondsoort',
                           'hoofdnaam2_grondsoort',
                           'bijmenging1_grondsoort',
                           'bijmenging1_hoeveelheid',
                           'bijmenging1_plaatselijk',
                           'bijmenging2_grondsoort',
                           'bijmenging3_grondsoort'))

        assert df.hoofdnaam1_grondsoort[4] == 'XZ'
        assert df.hoofdnaam2_grondsoort[4] == 'KL'
        assert df.bijmenging1_grondsoort[4] == 'GL'
        assert df.bijmenging1_hoeveelheid[2] == 'M'
        # mind that the column below is of dtype 'object'
        assert df.bijmenging1_plaatselijk[2] is False
        assert df.bijmenging2_grondsoort[5] == 'GL'
        assert df.bijmenging3_grondsoort[8] == 'NN'



# -*- coding: utf-8 -*-
"""Module grouping tests for the pydov.types.bodemmonster module."""

from pydov.types.bodemmonster import Bodemmonster
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestTypes

location_wfs_getfeature = 'tests/data/types/bodemmonster/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/bodemmonster/feature.xml'
location_dov_xml = 'tests/data/types/bodemmonster/bodemmonster.xml'


class TestBodemmonster(AbstractTestTypes):
    """Class grouping tests for the pydov.types.bodemmonster.Bodemmonster class."""

    datatype_class = Bodemmonster
    namespace = 'https://www.dov.vlaanderen.be/bodem'
    pkey_base = build_dov_url('data/bodemmonster/')

    field_names = [
            'pkey_bodemmonster', 'pkey_bodemlocatie', 'pkey_parent',
            'x', 'y', 'mv_mtaw', 'identificatie',
            'datum_monstername', 'tijdstip_monstername',
            'type', 'monstername_door', 'techniek',
            'condities', 'diepte_van_cm', 'diepte_tot_cm', 'labo'
        ]
    field_names_subtypes = []
    field_names_nosubtypes = [
            'pkey_bodemmonster', 'pkey_bodemlocatie', 'pkey_parent',
            'x', 'y', 'mv_mtaw', 'identificatie',
            'datum_monstername', 'tijdstip_monstername',
            'type', 'monstername_door', 'techniek',
            'condities', 'diepte_van_cm', 'diepte_tot_cm', 'labo'
    ]

    valid_returnfields = ('pkey_bodemmonster', 'identificatie')
    valid_returnfields_subtype = ('pkey_bodemmonster', 'identificatie')

    inexistent_field = 'onbestaand'

    def test_get_field_names_wrongreturnfields_nosubtypes(self):
        """Test the get_field_names method when disabling subtypes
        and including an otherwise valid return field.

        Test whether an InvalidFieldError is raised.

        Override because bodemmonster has no subtypes.
        """
        assert True



"""Module grouping tests for the pydov.util.location module."""
import pytest

from owslib.fes import (
    And,
    Or,
    Not,
)
from pydov.util.location import (
    Box,
    Point,
    Equals,
    Disjoint,
    Touches,
    Within,
    Intersects,
    WithinDistance,
    GmlObject
)
from owslib.etree import etree
from pydov.util.owsutil import set_geometry_column
from tests.abstract import clean_xml


class TestLocation(object):
    """Class grouping tests for the AbstractLocation subtypes."""

    def test_box(self):
        """Test the default Box type.

        Test whether the generated XML is correct.

        """
        box = Box(94720, 186910, 112220, 202870)
        xml = box.get_element()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:lowerCorner>94720.000000 186910.000000</gml:lowerCorner>'
            '<gml:upperCorner>112220.000000 202870.000000</gml:upperCorner>'
            '</gml:Envelope>')

    def test_box_wgs84(self):
        """Test the Box type with WGS84 coordinates.

        Test whether the generated XML is correct.

        """
        box = Box(3.6214, 50.9850, 3.8071, 51.1270, epsg=4326)
        xml = box.get_element()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#4326">'
            '<gml:lowerCorner>3.621400 50.985000</gml:lowerCorner>'
            '<gml:upperCorner>3.807100 51.127000</gml:upperCorner>'
            '</gml:Envelope>')

    def test_box_invalid(self):
        """Test the Box type with the wrong ordering of coordinates.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            Box(94720, 202870, 186910, 112220)

    def test_box_invalid_wgs84(self):
        """Test the Box type with the wrong ordering of WGS84 coordinates.

        Test whether a ValueError is raised.

        """
        with pytest.raises(ValueError):
            Box(50.9850, 3.6214, 3.8071, 51.1270, epsg=4326)

    def test_point(self):
        """Test the default Point type.

        Test whether the generated XML is correct.

        """
        point = Point(110680, 202030)
        xml = point.get_element()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:pos>110680.000000 202030.000000</gml:pos></gml:Point>')

    def test_point_wgs84(self):
        """Test the Point type with WGS84 coordinates.

        Test whether the generated XML is correct.

        """
        point = Point(3.8071, 51.1270, epsg=4326)
        xml = point.get_element()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#4326">'
            '<gml:pos>3.807100 51.127000</gml:pos></gml:Point>')

    def test_gmlobject_element(self):
        """Test the GmlObject type with an etree.Element.

        Test whether the returned XML is correct.

        """
        with open('tests/data/util/location/polygon_single_31370.gml',
                  'r') as gmlfile:
            gml = gmlfile.read()

            gml_element = etree.fromstring(gml.encode('utf8'))
            gml_element = gml_element.find(
                './/{http://www.opengis.net/gml}Polygon')

            gml_object = GmlObject(gml_element)

            assert clean_xml(etree.tostring(
                gml_object.get_element()).decode('utf8')) == clean_xml(
                '<gml:Polygon '
                'srsName="urn:ogc:def:crs:EPSG::31370"><gml:exterior><gml'
                ':LinearRing><gml:posList>108636.150020818 194960.844295764 '
                '108911.922161617 194291.111953824 109195.573506438 '
                '195118.42837622 108636.150020818 '
                '194960.844295764</gml:posList></gml:LinearRing></gml'
                ':exterior></gml:Polygon>')

    def test_gmlobject_bytes(self):
        """Test the GmlObject type with a GML string.

        Test whether the returned XML is correct.

        """
        with open('tests/data/util/location/polygon_single_31370.gml',
                  'r') as gmlfile:
            gml = gmlfile.read()

            gml_element = etree.fromstring(gml.encode('utf8'))
            gml_element = gml_element.find(
                './/{http://www.opengis.net/gml}Polygon')

            gml_object = GmlObject(etree.tostring(gml_element))

            assert clean_xml(etree.tostring(
                gml_object.get_element()).decode('utf8')) == clean_xml(
                '<gml:Polygon '
                'srsName="urn:ogc:def:crs:EPSG::31370"><gml:exterior><gml'
                ':LinearRing><gml:posList>108636.150020818 194960.844295764 '
                '108911.922161617 194291.111953824 109195.573506438 '
                '195118.42837622 108636.150020818 '
                '194960.844295764</gml:posList></gml:LinearRing></gml'
                ':exterior></gml:Polygon>')

    def test_gmlobject_string(self):
        """Test the GmlObject type with a GML string.

        Test whether the returned XML is correct.

        """
        with open('tests/data/util/location/polygon_single_31370.gml',
                  'r') as gmlfile:
            gml = gmlfile.read()

            gml_element = etree.fromstring(gml.encode('utf8'))
            gml_element = gml_element.find(
                './/{http://www.opengis.net/gml}Polygon')

            gml_object = GmlObject(etree.tostring(gml_element).decode('utf8'))

            assert clean_xml(etree.tostring(
                gml_object.get_element()).decode('utf8')) == clean_xml(
                '<gml:Polygon '
                'srsName="urn:ogc:def:crs:EPSG::31370"><gml:exterior><gml'
                ':LinearRing><gml:posList>108636.150020818 194960.844295764 '
                '108911.922161617 194291.111953824 109195.573506438 '
                '195118.42837622 108636.150020818 '
                '194960.844295764</gml:posList></gml:LinearRing></gml'
                ':exterior></gml:Polygon>')


class TestBinarySpatialFilters(object):
    """Class grouping tests for the AbstractBinarySpatialFilter subtypes."""

    def test_equals_point(self):
        """Test the Equals spatial filter with a Point location.

        Test whether the generated XML is correct.

        """
        equals = Equals(Point(150000, 150000))
        equals.set_geometry_column('geom')
        xml = equals.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Equals><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:pos>150000.000000 150000.000000</gml:pos></gml:Point>'
            '</ogc:Equals>')

    def test_equals_nogeom(self):
        """Test the Equals spatial filter without setting a geometry column.

        Test whether a RuntimeError is raised.

        """
        equals = Equals(Point(150000, 150000))

        with pytest.raises(RuntimeError):
            equals.toXML()

    def test_disjoint_box(self):
        """Test the Disjoint spatial filter with a Box location.

        Test whether the generated XML is correct.

        """
        disjoint = Disjoint(Box(94720, 186910, 112220, 202870))
        disjoint.set_geometry_column('geom')
        xml = disjoint.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Disjoint><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:lowerCorner>94720.000000 186910.000000</gml:lowerCorner>'
            '<gml:upperCorner>112220.000000 202870.000000</gml:upperCorner>'
            '</gml:Envelope></ogc:Disjoint>')

    def test_disjoint_nogeom(self):
        """Test the Disjoint spatial filter without setting a geometry column.

        Test whether a RuntimeError is raised.

        """
        disjoint = Disjoint(Point(150000, 150000))

        with pytest.raises(RuntimeError):
            disjoint.toXML()

    def test_touches_box(self):
        """Test the Touches spatial filter with a Box location.

        Test whether the generated XML is correct.

        """
        touches = Touches(Box(94720, 186910, 112220, 202870))
        touches.set_geometry_column('geom')
        xml = touches.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Touches><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:lowerCorner>94720.000000 186910.000000</gml:lowerCorner>'
            '<gml:upperCorner>112220.000000 202870.000000</gml:upperCorner>'
            '</gml:Envelope></ogc:Touches>')

    def test_touches_nogeom(self):
        """Test the Touches spatial filter without setting a geometry column.

        Test whether a RuntimeError is raised.

        """
        touches = Touches(Point(150000, 150000))

        with pytest.raises(RuntimeError):
            touches.toXML()

    def test_within_box(self):
        """Test the Within spatial filter with a Box location.

        Test whether the generated XML is correct.

        """
        within = Within(Box(94720, 186910, 112220, 202870))
        within.set_geometry_column('geom')
        xml = within.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Within><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:lowerCorner>94720.000000 186910.000000</gml:lowerCorner>'
            '<gml:upperCorner>112220.000000 202870.000000</gml:upperCorner>'
            '</gml:Envelope></ogc:Within>')

    def test_within_nogeom(self):
        """Test the Within spatial filter without setting a geometry column.

        Test whether a RuntimeError is raised.

        """
        within = Within(Box(94720, 186910, 112220, 202870))

        with pytest.raises(RuntimeError):
            within.toXML()

    def test_intersects_box(self):
        """Test the Intersects spatial filter with a Box location.

        Test whether the generated XML is correct.

        """
        intersects = Intersects(Box(94720, 186910, 112220, 202870))
        intersects.set_geometry_column('geom')
        xml = intersects.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Intersects><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:lowerCorner>94720.000000 186910.000000</gml:lowerCorner>'
            '<gml:upperCorner>112220.000000 202870.000000</gml:upperCorner>'
            '</gml:Envelope></ogc:Intersects>')

    def test_intersects_nogeom(self):
        """Test the Intersects spatial filter without setting a geometry
        column.

        Test whether a RuntimeError is raised.

        """
        intersects = Intersects(Box(94720, 186910, 112220, 202870))

        with pytest.raises(RuntimeError):
            intersects.toXML()


class TestLocationFilters(object):
    """Class grouping tests for the AbstractLocationFilter subtypes."""

    def test_withindistance_point(self):
        """Test the WithinDistance spatial filter with a Point location.

        Test whether the generated XML is correct.

        """
        withindistance = WithinDistance(Point(150000, 150000), 100)
        withindistance.set_geometry_column('geom')
        xml = withindistance.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:DWithin><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:pos>150000.000000 150000.000000</gml:pos></gml:Point>'
            '<gml:Distance units="meter">100.000000</gml:Distance>'
            '</ogc:DWithin>')

    def test_withindistance_point_named_args(self):
        """Test the WithinDistance spatial filter with a Point location.

        Test whether the generated XML is correct.

        """
        withindistance = WithinDistance(location=Point(150000, 150000),
                                        distance=100, distance_unit='meter')
        withindistance.set_geometry_column('geom')
        xml = withindistance.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:DWithin><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370">'
            '<gml:pos>150000.000000 150000.000000</gml:pos></gml:Point>'
            '<gml:Distance units="meter">100.000000</gml:Distance>'
            '</ogc:DWithin>')

    def test_withindistance_nogeom(self):
        """Test the WithinDistance spatial filter without setting a geometry
        column.

        Test whether a RuntimeError is raised.

        """
        withindistance = WithinDistance(Point(150000, 150000), 100)

        with pytest.raises(RuntimeError):
            withindistance.toXML()

    def test_withindistance_point_wgs84(self):
        """Test the WithinDistance spatial filter with a Point location
        using WGS84 coordinates.

        Test whether the generated XML is correct.

        """
        withindistance = WithinDistance(Point(51.1270, 3.8071, epsg=4326), 100)
        withindistance.set_geometry_column('geom')
        xml = withindistance.toXML()

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:DWithin><ogc:PropertyName>geom</ogc:PropertyName>'
            '<gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#4326">'
            '<gml:pos>51.127000 3.807100</gml:pos></gml:Point>'
            '<gml:Distance units="meter">100.000000</gml:Distance>'
            '</ogc:DWithin>')


class TestLocationFilterExpressions(object):
    """Class grouping tests for expressions with spatial filters."""

    def test_point_and_box(self):
        """Test a location filter expression using a Within(Box) and a
        WithinDistance(Point) filter.

        Test whether the generated XML is correct.

        """
        point_and_box = And([WithinDistance(Point(150000, 150000), 100),
                             Within(Box(94720, 186910, 112220, 202870))])
        xml = set_geometry_column(point_and_box, 'geom')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:And><ogc:DWithin><ogc:PropertyName>geom</ogc:PropertyName'
            '><gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370"><gml'
            ':pos>150000.000000 '
            '150000.000000</gml:pos></gml:Point><gml:Distance '
            'units="meter">100.000000</gml:Distance></ogc:DWithin><ogc'
            ':Within><ogc:PropertyName>geom</ogc:PropertyName><gml'
            ':Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370"><gml'
            ':lowerCorner>94720.000000 '
            '186910.000000</gml:lowerCorner><gml:upperCorner>112220.000000 '
            '202870.000000</gml:upperCorner></gml:Envelope></ogc:Within'
            '></ogc:And>')

    def test_box_or_box(self):
        """Test a location filter expression using an Intersects(Box) and a
        Within(Box) filter.

        Test whether the generated XML is correct.

        """
        box_or_box = Or([
            Intersects(Box(50.9850, 3.6214, 51.1270, 3.8071, epsg=4326)),
            Within(Box(94720, 186910, 112220, 202870))])
        xml = set_geometry_column(box_or_box, 'geom')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:Or><ogc:Intersects><ogc:PropertyName>geom</ogc'
            ':PropertyName><gml:Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#4326"><gml'
            ':lowerCorner>50.985000 '
            '3.621400</gml:lowerCorner><gml:upperCorner>51.127000 '
            '3.807100</gml:upperCorner></gml:Envelope></ogc:Intersects><ogc'
            ':Within><ogc:PropertyName>geom</ogc:PropertyName><gml:Envelope '
            'srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370"><gml '
            ':lowerCorner>94720.000000 '
            '186910.000000</gml:lowerCorner><gml:upperCorner>112220.000000 '
            '202870.000000</gml:upperCorner></gml:Envelope></ogc:Within'
            '></ogc:Or>')

    def test_recursive(self):
        """Test a location filter expression using a recursive expression
        with And(Not(WithinDistance(Point) filter.

        Test whether the generated XML is correct.

        """
        point_and_box = And([Not([WithinDistance(Point(150000, 150000), 100)]),
                             Within(Box(94720, 186910, 112220, 202870))])
        xml = set_geometry_column(point_and_box, 'geom')

        assert clean_xml(etree.tostring(xml).decode('utf8')) == clean_xml(
            '<ogc:And><ogc:Not><ogc:DWithin><ogc:PropertyName>geom</ogc'
            ':PropertyName><gml:Point srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370"><gml'
            ':pos>150000.000000 '
            '150000.000000</gml:pos></gml:Point><gml:Distance '
            'units="meter">100.000000</gml:Distance></ogc:DWithin></ogc:Not'
            '><ogc:Within><ogc:PropertyName>geom</ogc:PropertyName><gml'
            ':Envelope srsDimension="2" '
            'srsName="http://www.opengis.net/gml/srs/epsg.xml#31370"><gml'
            ':lowerCorner>94720.000000 '
            '186910.000000</gml:lowerCorner><gml:upperCorner>112220.000000 '
            '202870.000000</gml:upperCorner></gml:Envelope></ogc:Within'
            '></ogc:And>')



"""Module grouping tests for the search grondwaterfilter module."""
import datetime

from owslib.fes import PropertyIsEqualTo

from pydov.search.grondwaterfilter import GrondwaterFilterSearch
from pydov.types.grondwaterfilter import GrondwaterFilter
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/grondwaterfilter/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondwaterfilter/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondwaterfilter/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/grondwaterfilter/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondwaterfilter/feature.xml'
location_dov_xml = 'tests/data/types/grondwaterfilter/grondwaterfilter.xml'
location_xsd_base = 'tests/data/types/grondwaterfilter/xsd_*.xml'


class TestGrondwaterfilterSearch(AbstractTestSearch):

    search_instance = GrondwaterFilterSearch()
    datatype_class = GrondwaterFilter

    valid_query_single = PropertyIsEqualTo(propertyname='filterfiche',
                                           literal=build_dov_url(
                                               'data/filter/2003-004471'))

    inexistent_field = 'onbestaand'
    wfs_field = 'filternummer'
    xml_field = 'peil_mtaw'

    valid_returnfields = ('pkey_filter', 'filternummer')
    valid_returnfields_subtype = ('pkey_filter', 'filternummer', 'peil_mtaw')
    valid_returnfields_extra = ('pkey_filter', 'beheerder')

    df_default_columns = ['pkey_filter', 'pkey_grondwaterlocatie', 'gw_id',
                          'filternummer', 'filtertype', 'x', 'y',
                          'start_grondwaterlocatie_mtaw', 'mv_mtaw',
                          'gemeente', 'meetnet_code', 'aquifer_code',
                          'grondwaterlichaam_code', 'regime',
                          'diepte_onderkant_filter', 'lengte_filter',
                          'datum', 'tijdstip', 'peil_mtaw',
                          'betrouwbaarheid', 'methode', 'filterstatus',
                          'filtertoestand']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.datum.sort_values()[0] == datetime.date(2004, 4, 7)

    def test_search_xmlresolving(self, mp_get_schema,
                                 mp_remote_describefeaturetype,
                                 mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_filter', 'gw_id', 'filternummer',
                           'meetnet_code'))

        assert df.meetnet_code[0] == '8'



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV bodemmonster data."""
from ..types.bodemmonster import Bodemmonster
from .abstract import AbstractSearch


class BodemmonsterSearch(AbstractSearch):
    """Search class to retrieve information about bodemmonsters."""

    def __init__(self, objecttype=Bodemmonster):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemmonster type.
            Optional: defaults to the Bodemmonster type containing the fields
            described in the documentation.

        """
        super(BodemmonsterSearch, self).__init__('bodem:bodemmonsters',
                                                 objecttype)



"""Module for test for features without subtype values."""

from owslib.fes import PropertyIsEqualTo

from pydov.search.grondwaterfilter import GrondwaterFilterSearch
from pydov.util.dovutil import build_dov_url

location_md_metadata = 'tests/data/types/grondwaterfilter/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondwaterfilter/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondwaterfilter/wfsdescribefeaturetype.xml'
location_xsd_base = 'tests/data/types/grondwaterfilter/xsd_*.xml'

location_wfs_getfeature = \
    'tests/data/types/grondwaterfilter/wfsgetfeature_geenpeilmeting.xml'
location_wfs_feature = \
    'tests/data/types/grondwaterfilter/feature_geenpeilmeting.xml'
location_dov_xml = \
    'tests/data/types/grondwaterfilter/grondwaterfilter_geenpeilmeting.xml'


class TestSearchNoSubtype(object):
    def test_search_nosubtype(self, mp_wfs, mp_remote_md, mp_remote_fc,
                              mp_get_schema, mp_remote_describefeaturetype,
                              mp_remote_wfs_feature, mp_remote_xsd,
                              mp_dov_xml):
        """Test the search method with a result containing no elements from
        the subtype.

        Test whether the output dataframe contains the results, with the
        columns of the subtype set to NaN.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_remote_xsd : pytest.fixture
            Monkeypatch the call to get XSD schemas.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = GrondwaterFilterSearch().search(
            query=PropertyIsEqualTo(
                'pkey_filter',
                build_dov_url('data/filter/2007-011302.xml')
            )
        )

        assert len(df.pkey_filter) == 1

        assert df.datum.hasnans
        assert df.tijdstip.hasnans
        assert df.peil_mtaw.hasnans
        assert df.betrouwbaarheid.hasnans



"""Module grouping tests for the pydov.util.owsutil module."""
import copy
import os

import pytest

from pydov.util import dovutil

env_var = "PYDOV_BASE_URL"


@pytest.fixture
def pydov_base_url_environment():
    """Fixture for setting an environment variable with a different base_url.
    """
    old_environ = copy.deepcopy(os.environ)
    os.environ[env_var] = 'https://dov/'

    yield

    os.environ = old_environ


class TestDovutil(object):
    """Class grouping tests for the pydov.util.dovutil module."""

    def test_get_default_dov_base_url_slash(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('/geonetwork') == \
                   'https://www.dov.vlaanderen.be/geonetwork'

    def test_get_default_dov_base_url_multislash(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('/geonetwork/srv') == \
                   'https://www.dov.vlaanderen.be/geonetwork/srv'

    def test_get_default_dov_base_url_endslash(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('/geonetwork/') == \
                   'https://www.dov.vlaanderen.be/geonetwork/'

    def test_get_default_dov_base_url_noslash(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('geonetwork') == \
                   'https://www.dov.vlaanderen.be/geonetwork'

    def test_get_default_dov_base_url_noslash_multi(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('geonetwork/srv') == \
                   'https://www.dov.vlaanderen.be/geonetwork/srv'

    def test_get_default_dov_base_url_noslash_end(self):
        if env_var not in os.environ:
            assert dovutil.build_dov_url('geonetwork/') == \
                   'https://www.dov.vlaanderen.be/geonetwork/'

    def test_get_dov_base_url_slash(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('/geonetwork') == \
            'https://dov/geonetwork'

    def test_get_dov_base_url_multislash(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('/geonetwork/srv') == \
            'https://dov/geonetwork/srv'

    def test_get_dov_base_url_endslash(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('/geonetwork/') == \
            'https://dov/geonetwork/'

    def test_get_dov_base_url_noslash(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('geonetwork') == \
            'https://dov/geonetwork'

    def test_get_dov_base_url_noslash_multi(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('geonetwork/srv') == \
            'https://dov/geonetwork/srv'

    def test_get_dov_base_url_noslash_end(self, pydov_base_url_environment):
        assert env_var in os.environ
        assert dovutil.build_dov_url('geonetwork/') == \
            'https://dov/geonetwork/'



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemmonsters, including
subtypes."""
from pydov.types.fields import WfsField, XmlField

from .abstract import AbstractDovType


class Bodemmonster(AbstractDovType):
    """Class representing the DOV data type for bodemmonsters."""

    subtypes = []

    fields = [
        WfsField(name='pkey_bodemmonster', source_field='Bodemmonsterfiche',
                 datatype='string'),
        WfsField(name='pkey_bodemlocatie',
                 source_field='Bodemlocatiefiche',
                 datatype='string'),
        WfsField(name='pkey_parent',
                 source_field='Parentfiche',
                 datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='mv_mtaw', source_field='mv_mTAW', datatype='float'),
        WfsField(name='identificatie', source_field='Bodemmonster',
                 datatype='string'),
        WfsField(name='datum_monstername', source_field='Datum_monsterafname',
                 datatype='date'),
        XmlField(name='tijdstip_monstername',
                 source_xpath='/bodemmonster/tijdstip_monstername',
                 definition='Tijdstip waarop het monster werd genomen.',
                 datatype='string'),
        WfsField(name='type', source_field='Type', datatype='string'),
        WfsField(name='monstername_door', source_field='Monsterafname_door',
                 datatype='string'),
        WfsField(name='techniek', source_field='Techniek_monsterafname',
                 datatype='string'),
        WfsField(name='condities', source_field='Condities_monsterafname',
                 datatype='string'),
        WfsField(name='diepte_van_cm', source_field='Diepte_van',
                 datatype='float'),
        WfsField(name='diepte_tot_cm', source_field='Diepte_tot',
                 datatype='float'),
        WfsField(name='labo', source_field='Labo', datatype='string')
    ]

    pkey_fieldname = 'Bodemmonsterfiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemmonster, being a URI of the form
            `https://www.dov.vlaanderen.be/data/bodemmonster/<id>`.

        """
        super().__init__('bodemmonster', pkey)



# -*- coding: utf-8 -*-
"""Module implementing a local cache for downloaded XML files."""
import datetime
import gzip
import os
import re
import shutil
import tempfile
import warnings

from pydov.util.dovutil import build_dov_url, get_dov_xml
from pydov.util.errors import RemoteFetchError, XmlStaleWarning
from pydov.util.hooks import HookRunner


class AbstractCache(object):
    """Abstract base class for caching of downloaded XML files from DOV.

    Attributes
    ----------
    stale_on_error : bool, default to True
        Whether to return stale responses from the cache in case of a network
        error prevents downloading a fresh copy.
    """

    def __init__(self):
        """Initialisation."""
        self.stale_on_error = True

    def _get_remote(self, url, session=None):
        """Get the XML data by requesting it from the given URL.

        Parameters
        ----------
        url : str
            Permanent URL to a DOV object.
        session : requests.Session
            Session to use to perform HTTP requests for data. Defaults to None,
            which means a new session will be created for each request.

        Returns
        -------
        xml : bytes
            The raw XML data of this DOV object as bytes.

        """
        xml = get_dov_xml(url, session)
        HookRunner.execute_xml_downloaded(url.rstrip('.xml'))
        return xml

    def _emit_cache_hit(self, url):
        """Emit the XML cache hit event for all registered hooks.

        This notifies hooks that a valid XML document has been returned from
        the cache.

        Parameters
        ----------
        url : str
            Permanent URL to a DOV object.
        """
        HookRunner.execute_xml_cache_hit(url.rstrip('.xml'))

    def _emit_stale_hit(self, url):
        """Emit the XML stale hit event for all registered hooks.

        This notifies hooks that a stale XML document has been returned from
        the cache.

        Parameters
        ----------
        url : str
            Permanent URL to a DOV object.
        """
        HookRunner.execute_xml_stale_hit(url.rstrip('.xml'))

    def get(self, url, session=None):
        """Get the XML data for the DOV object referenced by the given URL.

        Because of parallel processing, this method will be called
        simultaneously from multiple threads. Make sure your implementation is
        threadsafe or uses locking.

        If a valid version exists in the cache, it will be loaded and
        returned. If no valid version exists, the XML will be downloaded
        from the DOV webservice, saved in the cache and returned.

        Parameters
        ----------
        url : str
            Permanent URL to a DOV object.
        session : requests.Session
            Session to use to perform HTTP requests for data. Defaults to None,
            which means a new session will be created for each request.

        Returns
        -------
        xml : bytes
            The raw XML data of this DOV object as bytes.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def clean(self):
        """Clean the cache by removing old records from the cache.

        Since during normal use the cache only grows by adding new objects and
        overwriting existing ones with a new version, you can use this
        function to clean the cache. It will remove all records older than
        the maximum age from the cache.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def remove(self):
        """Remove the entire cache."""
        raise NotImplementedError('This should be implemented in a subclass.')


class AbstractFileCache(AbstractCache):
    """Abstract class for filebased caching of downloaded XML files from
    DOV."""

    def __init__(self, max_age=datetime.timedelta(weeks=2), cachedir=None):
        """Initialisation.

        Set up the instance variables and create the cache directory if
        it does not exists already.

        Parameters
        ----------
        max_age : datetime.timedelta, optional
            The maximum age of a cached XML file to be valid. If the last
            modification date of the file is before this time, it will be
            redownloaded. Defaults to two weeks.
        cachedir : str, optional
            Path of the directory that will be used to save the cached XML
            files. Be sure to use a directory that will only be used for
            this PyDOV cache. Default to a temporary directory provided by
            the operating system.

        """
        super().__init__()

        if cachedir:
            self.cachedir = cachedir
        else:
            self.cachedir = os.path.join(tempfile.gettempdir(), 'pydov')
        self.max_age = max_age

        self._re_type_key = re.compile(
            build_dov_url('data/') + r'([^ /]+)/([^.]+)'
        )

        try:
            if not os.path.exists(self.cachedir):
                os.makedirs(self.cachedir)
        except Exception:
            pass

    def _get_filepath(self, datatype, key):
        """Get the location on disk where the object with given datatype and
        key is to be saved.

        Parameters
        ----------
        datatype : str
            Datatype of the DOV object.
        key : str
            Unique and permanent object key of the DOV object.

        Returns
        -------
        str
            Full absolute path on disk where the object is to be saved.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def _get_type_key_from_url(self, url):
        """Parse a DOV permalink and return the datatype and object key.

        Parameters
        ----------
        url : str
            Permanent URL to a DOV object.

        Returns
        -------
        datatype : str
            Datatype of the DOV object referred to by the URL.
        key : str
            Unique and permanent key of the instance of the DOV object
            referred to by the URL.

        """
        datatype = self._re_type_key.search(url)
        if datatype and len(datatype.groups()) > 1:
            return datatype.group(1), datatype.group(2)

    def _get_type_key_from_path(self, path):
        """Parse a filepath and return the datatype and object key.

        Parameters
        ----------
        path : str
            Full, absolute, path to a cached file.

        Returns
        -------
        datatype : str
            Datatype of the DOV object referred to by the URL.
        key : str
            Unique and permanent key of the instance of the DOV object
            referred to by the URL.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def _is_valid(self, datatype, key):
        """Check if a valid version of the given DOV object exists in the
        cache.

        A cached version is valid if it exists and the last modification
        time of the file is after the maximum age defined on initialisation.

        Parameters
        ----------
        datatype : str
            Datatype of the DOV object.
        key : str
            Unique and permanent object key of the DOV object.

        Returns
        -------
        bool
            True if a valid cached version exists, False otherwise.

        """
        filepath = self._get_filepath(datatype, key)
        if not os.path.exists(filepath):
            return False

        last_modification = datetime.datetime.fromtimestamp(
            os.path.getmtime(filepath))
        now = datetime.datetime.now()

        if (now - last_modification) > self.max_age:
            return False
        else:
            return True

    def _is_stale(self, datatype, key):
        """Check if a stale version of the given DOV object exists in the
        cache.

        A cached version is stale if it exists and the last modification
        time of the file is before the maximum age defined on initialisation.

        Parameters
        ----------
        datatype : str
            Datatype of the DOV object.
        key : str
            Unique and permanent object key of the DOV object.

        Returns
        -------
        bool
            True if a stale cached version exists, False otherwise.
        """
        if self._is_valid(datatype, key):
            return False

        filepath = self._get_filepath(datatype, key)
        return os.path.exists(filepath)

    def _load(self, datatype, key):
        """Read a cached version from disk.

        datatype : str
            Datatype of the DOV object.
        key : str
            Unique and permanent object key of the DOV object.

        Returns
        -------
        str (xml)
            XML string of the DOV object, loaded from the cache.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def _save(self, datatype, key, content):
        """Save the given content in the cache.

        Parameters
        ----------
        datatype : str
            Datatype of the DOV object to save.
        key : str
            Unique and permanent object key of the DOV object to save.
        content : bytes
            The raw XML data of this DOV object as bytes.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def get(self, url, session=None):
        datatype, key = self._get_type_key_from_url(url)

        data = HookRunner.execute_inject_xml_response(url)

        if data is not None:
            HookRunner.execute_xml_received(url, data)
            return data

        if self._is_valid(datatype, key):
            try:
                self._emit_cache_hit(url)
                data = self._load(datatype, key).encode('utf-8')

                HookRunner.execute_xml_received(url, data)
                return data
            except Exception:
                pass

        try:
            data = self._get_remote(url, session)
        except RemoteFetchError:
            if self.stale_on_error and self._is_stale(datatype, key):
                self._emit_stale_hit(url)
                warnings.warn((
                    "Failed to fetch remote XML document for "
                    "object '{}', using older stale version from cache. "
                    "Resulting dataframe will be out-of-date.".format(url)),
                    XmlStaleWarning)

                data = self._load(datatype, key).encode('utf-8')
                return data
            else:
                HookRunner.execute_xml_fetch_error(url)
                raise RemoteFetchError
        else:
            try:
                self._save(datatype, key, data)
            except Exception:
                pass

        return data

    def clean(self):
        """Clean the cache by removing old records from the cache.

        Since during normal use the cache only grows by adding new objects and
        overwriting existing ones with a new version, you can use this
        function to clean the cache. It will remove all records older than
        the maximum age from the cache.

        Note that this method is currently not called anywhere in the code,
        but it is provided as reference.

        """
        if os.path.exists(self.cachedir):
            for type in os.listdir(self.cachedir):
                for object in os.listdir(os.path.join(self.cachedir, type)):
                    datatype, key = self._get_type_key_from_path(
                        os.path.join(self.cachedir, type, object))
                    if not self._is_valid(datatype, key):
                        os.remove(
                            os.path.join(self.cachedir, datatype, object))

    def remove(self):
        """Remove the entire cache directory.

        Note that the default directory to save the cache is a temporary
        location provided by the operating system, and as a subsequence the
        OS will normally take care of its removal.

        Note that this method is currently not called anywhere in the code,
        but it is provided as reference.

        """
        if os.path.exists(self.cachedir):
            shutil.rmtree(self.cachedir)


class PlainTextFileCache(AbstractFileCache):
    """Class for plain text caching of downloaded XML files from DOV."""

    def _get_filepath(self, datatype, key):
        return os.path.join(self.cachedir, datatype, key + '.xml')

    def _get_type_key_from_path(self, path):
        key = os.path.basename(path).rstrip('.xml')
        datatype = os.path.dirname(path).split()[-1]
        return datatype, key

    def _save(self, datatype, key, content):
        filepath = self._get_filepath(datatype, key)
        folder = os.path.dirname(filepath)

        if not os.path.exists(folder):
            os.makedirs(folder)

        with open(filepath, 'w', encoding='utf-8') as f:
            f.write(content.decode('utf-8'))

    def _load(self, datatype, key):
        filepath = self._get_filepath(datatype, key)
        with open(filepath, 'r', encoding='utf-8') as f:
            return f.read()


class GzipTextFileCache(AbstractFileCache):
    """Class for GZipped text caching of downloaded XML files from DOV."""

    def _get_filepath(self, datatype, key):
        return os.path.join(self.cachedir, datatype, key + '.xml.gz')

    def _get_type_key_from_path(self, path):
        key = os.path.basename(path).rstrip('.xml.gz')
        datatype = os.path.dirname(path).split()[-1]
        return datatype, key

    def _save(self, datatype, key, content):
        filepath = self._get_filepath(datatype, key)
        folder = os.path.dirname(filepath)

        if not os.path.exists(folder):
            os.makedirs(folder)

        with gzip.open(filepath, 'wb') as f:
            f.write(content)

    def _load(self, datatype, key):
        filepath = self._get_filepath(datatype, key)
        with gzip.open(filepath, 'rb') as f:
            return f.read().decode('utf-8')



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV CPT data."""
from pydov.search.abstract import AbstractSearch
from pydov.types.sondering import Sondering


class SonderingSearch(AbstractSearch):
    """Search class to retrieve information about CPT measurements (
    Sonderingen)."""

    def __init__(self, objecttype=Sondering):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Sondering type.
            Optional: defaults to the Sondering type containing
            the fields described in the documentation.

        """
        super(SonderingSearch, self).__init__(
            'dov-pub:Sonderingen', objecttype)



# -*- coding: utf-8 -*-
"""Module grouping utility functions for DOV XML services."""
import os
import warnings

from owslib.etree import etree

from pydov.util.errors import RemoteFetchError, XmlParseError, XsdFetchWarning
from pydov.util.hooks import HookRunner
from pydov.util.net import SessionFactory


def build_dov_url(path):
    """Build the DOV url consisting of the fixed DOV base url, appended with
    the given path.

    Returns
    -------
    str
        The absolute DOV url.

    """
    if 'PYDOV_BASE_URL' in os.environ:
        base_url = os.environ['PYDOV_BASE_URL']
    else:
        base_url = 'https://www.dov.vlaanderen.be/'

    return base_url + path.lstrip('/')


def get_remote_url(url, session=None):
    """Request the URL from the remote service and return its contents.

    Parameters
    ----------
    url : str
        URL to download.
    session : requests.Session
        Session to use to perform HTTP requests for data. Defaults to None,
        which means a new session will be created for each request.

    Returns
    -------
    xml : bytes
        The raw XML data as bytes.

    """
    if session is None:
        session = SessionFactory.get_session()

    request = session.get(url)
    if request.status_code != 200:
        raise RemoteFetchError("Failed to fetch data at {}".format(url))

    request.encoding = 'utf-8'
    return request.text.encode('utf8')


def get_xsd_schema(url):
    """Request the XSD schema from DOV webservices and return it.

    Parameters
    ----------
    url : str
        URL of the XSD schema to download.

    Returns
    -------
    xml : bytes
        The raw XML data of this XSD schema as bytes.

    """
    response = HookRunner.execute_inject_meta_response(url)

    if response is None:
        try:
            response = get_remote_url(url)
        except RemoteFetchError:
            warnings.warn("Failed to fetch remote XSD schema, metadata will "
                          "be incomplete.", XsdFetchWarning)
            response = None

    HookRunner.execute_meta_received(url, response)

    return response


def get_dov_xml(url, session=None):
    """Request the XML from the remote DOV webservices and return it.

    Parameters
    ----------
    url : str
        URL of the DOV object to download.
    session : requests.Session
        Session to use to perform HTTP requests for data. Defaults to None,
        which means a new session will be created for each request.

    Returns
    -------
    xml : bytes
        The raw XML data of this DOV object as bytes.

    """
    response = HookRunner.execute_inject_xml_response(url)

    if response is None:
        response = get_remote_url(url, session)

    HookRunner.execute_xml_received(url, response)

    return response


def parse_dov_xml(xml_data):
    """Parse the given XML data into an ElementTree.

    Parameters
    ----------
    xml_data : bytes
        The raw XML data of a DOV object as bytes.

    Returns
    -------
    tree : etree.ElementTree
        Parsed XML tree of the DOV object.

    """
    try:
        parser = etree.XMLParser(ns_clean=True, recover=True)
    except TypeError:
        parser = etree.XMLParser()

    try:
        tree = etree.fromstring(xml_data, parser=parser)
        return tree
    except Exception:
        raise XmlParseError("Failed to parse XML record.")



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV
bodemobservatie data."""
from ..types.bodemobservatie import Bodemobservatie
from .abstract import AbstractSearch


class BodemobservatieSearch(AbstractSearch):
    """Search class to retrieve information about bodemobservaties."""

    def __init__(self, objecttype=Bodemobservatie):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the Bodemobservatie type.
            Optional: defaults to the Bodemobservatie type containing the
            fields described in the documentation.

        """
        super(BodemobservatieSearch, self).__init__('bodem:bodemobservaties',
                                                    objecttype)



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemsites, including
subtypes."""
from pydov.types.fields import WfsField, XmlField

from .abstract import AbstractDovType


class Bodemsite(AbstractDovType):
    """Class representing the DOV data type for bodemsites."""

    subtypes = []

    fields = [
        WfsField(name='pkey_bodemsite', source_field='Bodemsitefiche',
                 datatype='string'),
        WfsField(name='naam', source_field='Naam', datatype='string'),
        WfsField(name='waarnemingsdatum', source_field='Datum',
                 datatype='date'),
        WfsField(name='beschrijving', source_field='Beschrijving',
                 datatype='string'),
        XmlField(name='invoerdatum',
                 source_xpath='/bodemsite/invoerdatum',
                 definition='Datum van invoer van de bodemsite.',
                 datatype='date')
    ]

    pkey_fieldname = 'Bodemsitefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemsite, being a URI of the form
            `https://www.dov.vlaanderen.be/data/bodemsite/<id>`.

        """
        super().__init__('bodemsite', pkey)



import pytest
from owslib.fes import PropertyIsEqualTo

from pydov.search.grondwaterfilter import GrondwaterFilterSearch
from pydov.types.abstract import AbstractDovSubType
from pydov.types.fields import XmlField
from pydov.types.grondwaterfilter import GrondwaterFilter
from pydov.util.dovutil import build_dov_url

location_md_metadata = 'tests/data/types/grondwaterfilter/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondwaterfilter/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondwaterfilter/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/grondwaterfilter/wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondwaterfilter/feature.xml'
location_dov_xml = 'tests/data/types/grondwaterfilter/grondwaterfilter.xml'
location_xsd_base = 'tests/data/types/grondwaterfilter/xsd_*.xml'


class MyGrondwaterFilter(GrondwaterFilter):

    fields = GrondwaterFilter.extend_fields([
        XmlField(name='grondwatersysteem',
                 source_xpath='/filter/ligging/grondwatersysteem',
                 definition='Grondwatersysteem waarin de filter hangt.',
                 datatype='string')
    ])


class MyWrongGrondwaterFilter(GrondwaterFilter):

    fields = GrondwaterFilter.extend_fields([
        {'name': 'grondwatersysteem',
         'source': 'xml',
         'sourcefield': '/filter/ligging/grondwatersysteem',
         'definition': 'Grondwatersysteem waarin de filter hangt.',
         'type': 'string',
         'notnull': False
         }
    ])


class MyFilterOpbouw(AbstractDovSubType):

    rootpath = './/filter/opbouw/onderdeel'

    fields = [
        XmlField(name='opbouw_van',
                 source_xpath='/van',
                 definition='Opbouw van',
                 datatype='float'),
        XmlField(name='opbouw_tot',
                 source_xpath='/tot',
                 definition='Opbouw tot',
                 datatype='float'),
        XmlField(name='opbouw_element',
                 source_xpath='/filterelement',
                 definition='Opbouw element',
                 datatype='string',
                 notnull=False)
    ]


class MyGrondwaterFilterOpbouw(GrondwaterFilter):

    subtypes = [MyFilterOpbouw]


class TestMyWrongGrondwaterFilter(object):
    """Class grouping tests for the MyWrongGrondwaterFilter custom type."""
    def test_get_fields(self):
        """Test the get_fields method.

        Test whether a RuntimeError is raised.

        """
        fs = GrondwaterFilterSearch(objecttype=MyWrongGrondwaterFilter)

        with pytest.raises(RuntimeError):
            fs.get_fields()

    def test_search(self, mp_wfs, mp_remote_describefeaturetype,
                    mp_remote_md, mp_remote_fc, mp_remote_wfs_feature,
                    mp_dov_xml):
        """Test the search method.

        Test whether a RuntimeError is raised.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        fs = GrondwaterFilterSearch(objecttype=MyWrongGrondwaterFilter)

        with pytest.raises(RuntimeError):
            fs.search(query=PropertyIsEqualTo(
                propertyname='filterfiche',
                literal=build_dov_url('data/filter/2003-004471')))


class TestMyGrondwaterFilter(object):
    """Class grouping tests for the MyGrondwaterFilter custom type."""
    def test_get_fields(self):
        """Test the get_fields method.

        Test whether the extra field is available in the output of the
        get_fields metadata.

        """
        fs = GrondwaterFilterSearch(objecttype=MyGrondwaterFilter)
        fields = fs.get_fields()

        assert 'grondwatersysteem' in fields

    def test_search(self, mp_wfs, mp_remote_describefeaturetype,
                    mp_remote_md, mp_remote_fc, mp_remote_wfs_feature,
                    mp_dov_xml):
        """Test the search method.

        Test whether the extra fields from the custom type are resolved into
        data in the result dataframe.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        fs = GrondwaterFilterSearch(objecttype=MyGrondwaterFilter)

        df = fs.search(query=PropertyIsEqualTo(
            propertyname='filterfiche',
            literal=build_dov_url('data/filter/2003-004471')))

        assert 'grondwatersysteem' in df
        assert df.iloc[0].grondwatersysteem == 'Centraal Vlaams Systeem'


class TestMyGrondwaterFilterOpbouw(object):
    """Class grouping tests for the MyGrondwaterFilterOpbouw and
    MyFilterOpbouw custom type."""
    def test_get_fields(self):
        """Test the get_fields method.

        Test whether the extra field is available in the output of the
        get_fields metadata.

        """
        fs = GrondwaterFilterSearch(objecttype=MyGrondwaterFilterOpbouw)
        fields = fs.get_fields()

        assert 'datum' not in fields
        assert 'peil_mtaw' not in fields

        assert 'opbouw_van' in fields
        assert 'opbouw_tot' in fields
        assert 'opbouw_element' in fields

    def test_search(self, mp_wfs, mp_remote_describefeaturetype,
                    mp_remote_md, mp_remote_fc, mp_remote_wfs_feature,
                    mp_dov_xml):
        """Test the search method.

        Test whether the extra fields from the custom type are resolved into
        data in the result dataframe.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        fs = GrondwaterFilterSearch(objecttype=MyGrondwaterFilterOpbouw)

        df = fs.search(query=PropertyIsEqualTo(
            propertyname='filterfiche',
            literal=build_dov_url('data/filter/2003-004471')))

        assert 'opbouw_van' in df
        assert 'opbouw_tot' in df
        assert 'opbouw_element' in df

        assert df.iloc[-1].opbouw_van == 2.5
        assert df.iloc[-1].opbouw_tot == 2.7
        assert df.iloc[-1].opbouw_element == 'zandvang'



"""Module grouping tests for the search grondwatermonster module."""
import datetime

from owslib.fes import PropertyIsEqualTo

from pydov.search.grondwatermonster import GrondwaterMonsterSearch
from pydov.types.grondwatermonster import GrondwaterMonster
from pydov.util.dovutil import build_dov_url
from tests.abstract import AbstractTestSearch

location_md_metadata = 'tests/data/types/grondwatermonster/md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/grondwatermonster/fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/grondwatermonster/wfsdescribefeaturetype.xml'
location_wfs_getfeature = 'tests/data/types/grondwatermonster/' \
    'wfsgetfeature.xml'
location_wfs_feature = 'tests/data/types/grondwatermonster/feature.xml'
location_dov_xml = 'tests/data/types/grondwatermonster/grondwatermonster.xml'
location_xsd_base = 'tests/data/types/grondwatermonster/xsd_*.xml'


class TestGrondwaterMonsterSearch(AbstractTestSearch):

    search_instance = GrondwaterMonsterSearch()
    datatype_class = GrondwaterMonster

    valid_query_single = PropertyIsEqualTo(
        propertyname='grondwatermonsterfiche',
        literal=build_dov_url('data/watermonster/2006-115684'))

    inexistent_field = 'onbestaand'
    wfs_field = 'kationen'
    xml_field = 'eenheid'

    valid_returnfields = ('pkey_grondwatermonster', 'datum_monstername')
    valid_returnfields_subtype = (
        'pkey_grondwatermonster', 'datum_monstername', 'eenheid')
    valid_returnfields_extra = ('pkey_grondwatermonster', 'kationen')

    df_default_columns = [
        'pkey_grondwatermonster', 'grondwatermonsternummer',
        'pkey_grondwaterlocatie', 'gw_id', 'pkey_filter',
        'filternummer', 'x', 'y', 'start_grondwaterlocatie_mtaw',
        'gemeente', 'datum_monstername', 'parametergroep',
        'parameter', 'detectie', 'waarde', 'eenheid', 'veld_labo']

    def test_search_date(self, mp_wfs, mp_get_schema,
                         mp_remote_describefeaturetype, mp_remote_md,
                         mp_remote_fc, mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with only the query parameter.

        Test whether the result is correct.

        Parameters
        ----------
        mp_wfs : pytest.fixture
            Monkeypatch the call to the remote GetCapabilities request.
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_md : pytest.fixture
            Monkeypatch the call to get the remote metadata.
        mp_remote_fc : pytest.fixture
            Monkeypatch the call to get the remote feature catalogue.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single)

        # specific test for the Zulu time wfs 1.1.0 issue
        assert df.datum_monstername.sort_values()[0] == datetime.date(
            2006, 5, 19)



import copy
import datetime

import pytest
from owslib.etree import etree
from owslib.fes import PropertyIsEqualTo

import pydov
from pydov.search.boring import BoringSearch
from pydov.util.hooks import (AbstractInjectHook, AbstractReadHook, Hooks,
                              SimpleStatusHook)
from tests.abstract import ServiceCheck


@pytest.fixture
def test_hook_count():
    """PyTest fixture temporarily disabling default hooks and installing
    HookCounter."""
    orig_hooks = pydov.hooks

    pydov.hooks = Hooks(
        (HookCounter(),)
    )
    yield

    pydov.hooks = orig_hooks


@pytest.fixture
def test_hook_types():
    """PyTest fixture temporarily disabling default hooks and installing
    HookTester."""
    orig_hooks = pydov.hooks

    pydov.hooks = Hooks(
        (HookTypeTester(),)
    )
    yield

    pydov.hooks = orig_hooks


@pytest.fixture
def test_hook_inject():
    """PyTest fixture temporarily disabling default hooks and installing
    HookInjecter."""
    orig_hooks = pydov.hooks

    pydov.hooks = Hooks(
        (HookInjecter(),)
    )
    yield

    pydov.hooks = orig_hooks


class HookCounter(AbstractReadHook, AbstractInjectHook):
    """Hook implementation for testing purposes, counting all event calls."""

    def __init__(self):
        self.count_meta_received = 0
        self.count_inject_meta_response = 0
        self.count_wfs_search_init = 0
        self.count_wfs_search_result = 0
        self.count_wfs_search_result_received = 0
        self.count_inject_wfs_getfeature_response = 0
        self.count_xml_received = 0
        self.count_inject_xml_response = 0
        self.count_xml_cache_hit = 0
        self.count_xml_stale_hit = 0
        self.count_xml_fetch_error = 0
        self.count_xml_downloaded = 0

    def meta_received(self, url, response):
        self.count_meta_received += 1

    def inject_meta_response(self, url):
        self.count_inject_meta_response += 1

    def wfs_search_init(self, typename):
        self.count_wfs_search_init += 1

    def wfs_search_result(self, number_of_results):
        self.count_wfs_search_result += 1

    def wfs_search_result_received(self, query, features):
        self.count_wfs_search_result_received += 1

    def inject_wfs_getfeature_response(self, query):
        self.count_inject_wfs_getfeature_response += 1

    def xml_received(self, pkey_object, xml):
        self.count_xml_received += 1

    def inject_xml_response(self, pkey_object):
        self.count_inject_xml_response += 1

    def xml_cache_hit(self, pkey_object):
        self.count_xml_cache_hit += 1

    def xml_stale_hit(self, pkey_object):
        self.count_xml_stale_hit += 1

    def xml_downloaded(self, pkey_object):
        self.count_xml_downloaded += 1

    def xml_fetch_error(self, pkey_object):
        self.count_xml_fetch_error += 1


class HookTypeTester(AbstractReadHook, AbstractInjectHook):
    """Hook implementation for testing purposes, testing arguments of all
    event calls."""

    def meta_received(self, url, response):
        assert url is not None
        assert response is not None
        assert isinstance(url, str)
        assert isinstance(response, bytes)

    def inject_meta_response(self, url):
        assert url is not None
        assert isinstance(url, str)

    def wfs_search_init(self, typename):
        assert typename is not None
        assert isinstance(typename, str)

    def wfs_search_result(self, number_of_results):
        assert number_of_results is not None
        assert number_of_results > 0
        assert isinstance(number_of_results, int)

    def wfs_search_result_received(self, query, features):
        assert query is not None
        assert features is not None

    def inject_wfs_getfeature_response(self, query):
        assert query is not None

    def xml_received(self, pkey_object, xml):
        assert pkey_object is not None
        assert xml is not None
        assert isinstance(pkey_object, str)
        assert isinstance(xml, bytes)

    def inject_xml_response(self, pkey_object):
        assert pkey_object is not None
        assert isinstance(pkey_object, str)

    def xml_cache_hit(self, pkey_object):
        assert pkey_object is not None
        assert isinstance(pkey_object, str)

    def xml_stale_hit(self, pkey_object):
        assert pkey_object is not None
        assert isinstance(pkey_object, str)

    def xml_downloaded(self, pkey_object):
        assert pkey_object is not None
        assert isinstance(pkey_object, str)

    def xml_fetch_error(self, pkey_object):
        assert pkey_object is not None
        assert isinstance(pkey_object, str)


class HookInjecter(AbstractReadHook, AbstractInjectHook):
    """Hook implementation for testing purposes, testing response injection."""

    def __init__(self):
        self.wfs_features = None
        self.dov_xml = None

    def wfs_search_result_received(self, query, features):
        """Save the WFS GetFeature response in order to adjust and inject
        later."""
        self.wfs_features = features

    def inject_wfs_getfeature_response(self, query):
        """Adapt the previously saved WFS GetFeature response and inject the
        adapted version."""
        if self.wfs_features is not None:
            tree = copy.deepcopy(self.wfs_features)
            gemeentes = tree.findall(
                './/{http://dov.vlaanderen.be/ocdov/dov-pub}Boringen/'
                '{http://dov.vlaanderen.be/ocdov/dov-pub}gemeente')
            for g in gemeentes:
                g.text = 'Bevergem'
            return etree.tostring(tree, encoding='utf-8')

    def xml_received(self, pkey_object, xml):
        """Save the DOV Xml response in order to adjust and inject later."""
        self.dov_xml = xml

    def inject_xml_response(self, pkey_object):
        """Adapt the previously saved DOV XML response and inject the adapted
        version."""
        if self.dov_xml is not None:
            tree = etree.fromstring(self.dov_xml)
            boormethoden = tree.findall(
                './/boring/details/boormethode/methode')
            for b in boormethoden:
                b.text = 'De Pypere 106 T'
            return etree.tostring(tree, encoding='utf-8')


class TestHookCount(object):
    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_wfs_only(self, test_hook_count):
        """Test the search method providing both a location and a query.

        Test whether a dataframe is returned.

        Parameters
        ----------
        test_hook_count : pytest.fixture
            Fixture removing default hooks and installing HookCounter.

        """
        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()
        boringsearch.search(
            query=query, return_fields=('pkey_boring', 'x', 'y'))

        assert pydov.hooks[0].count_wfs_search_init == 1
        assert pydov.hooks[0].count_wfs_search_result == 1
        assert pydov.hooks[0].count_wfs_search_result_received == 1
        assert pydov.hooks[0].count_inject_wfs_getfeature_response == 1

        assert pydov.hooks[0].count_xml_received == 0
        assert pydov.hooks[0].count_inject_xml_response == 0
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 0
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_wfs_and_xml_nocache(self, test_hook_count, nocache):
        """Test the search method providing both a location and a query.

        Test whether a dataframe is returned.

        Parameters
        ----------
        test_hook_count : pytest.fixture
            Fixture removing default hooks and installing HookCounter.
        nocache : pytest.fixture
            Fixture temporarily disabling caching.

        """
        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()
        boringsearch.search(
            query=query, return_fields=('pkey_boring', 'mv_mtaw'))

        assert pydov.hooks[0].count_wfs_search_init == 1
        assert pydov.hooks[0].count_wfs_search_result == 1
        assert pydov.hooks[0].count_wfs_search_result_received == 1
        assert pydov.hooks[0].count_inject_wfs_getfeature_response == 1

        assert pydov.hooks[0].count_xml_received == 1
        assert pydov.hooks[0].count_inject_xml_response == 1
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 1
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0

        boringsearch.search(
            query=query, return_fields=('pkey_boring', 'mv_mtaw'))

        assert pydov.hooks[0].count_wfs_search_init == 2
        assert pydov.hooks[0].count_wfs_search_result == 2
        assert pydov.hooks[0].count_wfs_search_result_received == 2
        assert pydov.hooks[0].count_inject_wfs_getfeature_response == 2

        assert pydov.hooks[0].count_xml_received == 2
        assert pydov.hooks[0].count_inject_xml_response == 2
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 2
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    @pytest.mark.parametrize('plaintext_cache',
                             [[datetime.timedelta(minutes=15)]],
                             indirect=['plaintext_cache'])
    def test_wfs_and_xml_cache(self, test_hook_count, plaintext_cache):
        """Test the search method providing both a location and a query.

        Test whether a dataframe is returned.

        Parameters
        ----------
        test_hook_count : pytest.fixture
            Fixture removing default hooks and installing HookCounter.
        plaintext_cache : pytest.fixture
            Fixture temporarily setting up a testcache with max_age of 1
            second.

        """
        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()
        boringsearch.search(
            query=query, return_fields=('pkey_boring', 'mv_mtaw'))

        assert pydov.hooks[0].count_wfs_search_init == 1
        assert pydov.hooks[0].count_wfs_search_result == 1
        assert pydov.hooks[0].count_wfs_search_result_received == 1
        assert pydov.hooks[0].count_inject_wfs_getfeature_response == 1

        assert pydov.hooks[0].count_xml_received == 1
        assert pydov.hooks[0].count_inject_xml_response == 2
        assert pydov.hooks[0].count_xml_cache_hit == 0
        assert pydov.hooks[0].count_xml_downloaded == 1
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0

        boringsearch.search(
            query=query, return_fields=('pkey_boring', 'mv_mtaw'))

        assert pydov.hooks[0].count_wfs_search_init == 2
        assert pydov.hooks[0].count_wfs_search_result == 2
        assert pydov.hooks[0].count_wfs_search_result_received == 2
        assert pydov.hooks[0].count_inject_wfs_getfeature_response == 2

        assert pydov.hooks[0].count_xml_received == 2
        assert pydov.hooks[0].count_inject_xml_response == 3
        assert pydov.hooks[0].count_xml_cache_hit == 1
        assert pydov.hooks[0].count_xml_downloaded == 1
        assert pydov.hooks[0].count_xml_stale_hit == 0
        assert pydov.hooks[0].count_xml_fetch_error == 0

        assert pydov.hooks[0].count_meta_received > 0
        assert pydov.hooks[0].count_inject_meta_response > 0


class TestHookTypes(object):
    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_default_hooks(self, nocache):
        """Test the default hooks by performing a simple search.

        Test whether no exceptions are raised.

        Parameters
        ----------
        nocache : pytest.fixture
            Fixture temporarily disabling caching.

        """
        pydov.hooks = Hooks(
            (SimpleStatusHook(),)
        )

        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()
        boringsearch.search(query=query)

    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_hooks(self, test_hook_types):
        """Test the argument types of the hook events.

        Parameters
        ----------
        test_hook_types : pytest.fixture
            Fixture removing default hooks and installing HookTester.

        """
        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()
        boringsearch.search(query=query)


class TestHookInject(object):
    @pytest.mark.online
    @pytest.mark.skipif(not ServiceCheck.service_ok(),
                        reason="DOV service is unreachable")
    def test_hooks_inject(self, test_hook_inject):
        """Test the of the hook inject events.

        Test whether the requests are intercepted correctly.

        Parameters
        ----------
        test_hook_types : pytest.fixture
            Fixture removing default hooks and installing HookTester.

        """
        query = PropertyIsEqualTo(propertyname='boornummer',
                                  literal='GEO-04/169-BNo-B1')

        boringsearch = BoringSearch()

        boringsearch.search(query=query)
        df = boringsearch.search(query=query)

        assert df.iloc[0].gemeente == 'Bevergem'
        assert df.iloc[0].boormethode == 'De Pypere 106 T'



"""Module grouping tests for the interpretaties search module."""

import pandas as pd
from owslib.fes import PropertyIsEqualTo

from pydov.search.interpretaties import LithologischeBeschrijvingenSearch
from pydov.types.interpretaties import LithologischeBeschrijvingen
from tests.abstract import AbstractTestSearch

location_md_metadata = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'md_metadata.xml'
location_fc_featurecatalogue = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'fc_featurecatalogue.xml'
location_wfs_describefeaturetype = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'wfsdescribefeaturetype.xml'
location_wfs_getfeature = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/' \
    'wfsgetfeature.xml'
location_wfs_feature = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/feature.xml'
location_dov_xml = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen' \
    '/lithologische_beschrijvingen.xml'
location_xsd_base = \
    'tests/data/types/interpretaties/lithologische_beschrijvingen/xsd_*.xml'


class TestLithologischeBeschrijvingenSearch(AbstractTestSearch):

    search_instance = LithologischeBeschrijvingenSearch()
    datatype_class = LithologischeBeschrijvingen

    valid_query_single = PropertyIsEqualTo(propertyname='Proefnummer',
                                           literal='kb15d28w-B345')

    inexistent_field = 'onbestaand'
    wfs_field = 'Proefnummer'
    xml_field = 'beschrijving'

    valid_returnfields = ('pkey_interpretatie',
                          'betrouwbaarheid_interpretatie')
    valid_returnfields_subtype = (
        'pkey_interpretatie', 'diepte_laag_van', 'diepte_laag_tot')
    valid_returnfields_extra = ('pkey_interpretatie', 'gemeente')

    df_default_columns = [
        'pkey_interpretatie', 'pkey_boring',
        'betrouwbaarheid_interpretatie', 'x', 'y',
        'start_interpretatie_mtaw',
        'diepte_laag_van', 'diepte_laag_tot', 'beschrijving']

    def test_search_customreturnfields(self, mp_get_schema,
                                       mp_remote_describefeaturetype,
                                       mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with custom return fields.

        Test whether the output dataframe is correct.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType .
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'pkey_boring')
        )

        assert list(df) == ['pkey_interpretatie', 'pkey_boring']

        assert not pd.isnull(df.pkey_boring[0])

    def test_search_xml_resolve(self, mp_get_schema,
                                mp_remote_describefeaturetype,
                                mp_remote_wfs_feature, mp_dov_xml):
        """Test the search method with return fields from XML but not from a
        subtype.

        Test whether the output dataframe contains the resolved XML data.

        Parameters
        ----------
        mp_get_schema : pytest.fixture
            Monkeypatch the call to a remote OWSLib schema.
        mp_remote_describefeaturetype : pytest.fixture
            Monkeypatch the call to a remote DescribeFeatureType.
        mp_remote_wfs_feature : pytest.fixture
            Monkeypatch the call to get WFS features.
        mp_dov_xml : pytest.fixture
            Monkeypatch the call to get the remote XML data.

        """
        df = self.search_instance.search(
            query=self.valid_query_single,
            return_fields=('pkey_interpretatie', 'diepte_laag_tot'))

        assert df.diepte_laag_tot[0] == 1



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve permit data."""
import pandas as pd

from pydov.search.abstract import AbstractSearch
from pydov.types.grondwatervergunning import GrondwaterVergunning


class GrondwaterVergunningSearch(AbstractSearch):
    """Search class to retrieve information about Gw permits (
    all permits: current and historical)."""

    def __init__(self, objecttype=GrondwaterVergunning):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GrondwaterVergunning type.
            Optional: defaults to the GrondwaterVergunning type containing
            the fields described in the documentation.

        """
        super(GrondwaterVergunningSearch, self).__init__(
            'gw_vergunningen:alle_verg', objecttype)

    def search(self, location=None, query=None, sort_by=None,
               return_fields=None, max_features=None):
        fts = self._search(location=location, query=query, sort_by=sort_by,
                           return_fields=return_fields,
                           max_features=max_features)

        gw_vergunningen = self._type.from_wfs(fts, self._wfs_namespace)

        df = pd.DataFrame(
            data=self._type.to_df_array(gw_vergunningen, return_fields),
            columns=self._type.get_field_names(return_fields))
        return df



# -*- coding: utf-8 -*-
"""Module grouping classes for location based filters used for searching.

This module is designed to comply with the WFS 1.1.0 standard, implying
Filter Encoding 1.1 and GML 3.1.1.

"""
import os
from collections import OrderedDict
from distutils.version import LooseVersion
from io import BytesIO

from owslib.etree import etree
from owslib.fes import Or


class AbstractLocation(object):
    """Abstract base class for location types (f.ex. point, box, polygon).

    Locations are GML elements, for inclusion in the WFS GetFeature request.
    As described in the Filter Encoding 1.1 standard, locations are expressed
    using GML 3.1.1.

    The initialisation should require all necessary parameters to construct
    a valid location of this type: i.e. all locations should be valid after
    initialisation.

    """

    def get_element(self):
        """Return the GML representation of the location.

        Returns
        -------
        etree.Element
            XML element of the GML representation of this location.
        """
        raise NotImplementedError('This should be implemented in a subclass.')


class AbstractLocationFilter(object):
    """Abstract base class for location filters (f.ex. within, dwithin).

    Location filters are ogc:SpatialOpsType elements, for inclusion in
    the WFS GetFeature request.

    The initialisation should accept at least a (subclass of)
    AbstractLocation and require all additional parameters that are required
    for a valid location filter of this type.

    One exception is the name of the geometry column to query: this is set
    automatically upon building the WFS GetFeature request using the
    `set_geometry_column` method.

    """

    def set_geometry_column(self, geometry_column):
        """Set the name of the geometry column to query.

        Parameters
        ----------
        geometry_column : str
            The name of the geometry column to query.

        """
        raise NotImplementedError('This should be implemented in a subclass.')

    def toXML(self):
        """Return the XML representation of the location filter.

        Should raise a RuntimeError when called before the geometry column
        is set through `set_geometry_column`: location filters without the
        geometry column name are invalid.

        Returns
        -------
        etree.Element
            XML element of this location filter.

        Raises
        ------
        RuntimeError
            When called before the geometry column name is set: location
            filters without the geometry column name are invalid.

        """
        raise NotImplementedError('This should be implemented in a subclass.')


class AbstractBinarySpatialFilter(AbstractLocationFilter):
    """Class representing a binary spatial filter.

    Binary spatial filters are ogc:BinarySpatialOpType elements, i.e. one of
    Equals, Disjoint, Touches, Within, Overlaps, Crosses, Intersects or
    Contains.

    """

    def __init__(self, type, location):
        """Initialise a Binary spatial filter.

        Parameters
        ----------
        type : str
            Type of this filter: one of Equals, Disjoint, Touches, Within,
            Overlaps, Crosses, Intersects or Contains.
        location : AbstractLocation
            An instance of a location to use as location for the spatial
            filter.

        """
        self.location = location
        self.geom_column = ''

        self.element = etree.Element(
            '{{http://www.opengis.net/ogc}}{}'.format(type))

        geom = etree.Element('{http://www.opengis.net/ogc}PropertyName')
        geom.text = self.geom_column

        self.element.append(geom)
        self.element.append(location.get_element())

    def set_geometry_column(self, geometry_column):
        self.geom_column = geometry_column
        geom = self.element.find('.//{http://www.opengis.net/ogc}PropertyName')
        geom.text = geometry_column

    def toXML(self):
        if self.geom_column == '':
            raise RuntimeError('Geometry column has not been set. Use '
                               '"set_geometry_column" to set it.')
        return self.element


class Box(AbstractLocation):
    """Class representing a box location, also known as bounding box,
    envelope, extent."""

    def __init__(self, minx, miny, maxx, maxy, epsg=31370):
        """Initialise a Box.

        To initialise a Box using GPS coordinates in decimal degrees,
        use epsg=4326 and enter the longitude range as `minx` and `maxx` and
        the latitude range as `miny` and `maxy`.

        Parameters
        ----------
        minx : float
            X coordinate of the lower left corner of the box.
        miny : float
            Y coordinate of the lower left corner of the box.
        maxx : float
            X coordinate of the upper right corner of the box.
        maxy : float
            Y coordinate of the upper right corner of the box.
        epsg : int, optional
            EPSG code of the coordinate reference system (CRS) of the
            coordinates specified in `minx`, `miny`, `maxx`, `maxy`. Defaults
            to 31370 (Belgian Lambert72).

        Raises
        ------
        ValueError
            If `maxx` is lower than or equal to `minx`.
            If `maxy` is lower than or equal to `miny`.

        """
        if maxx <= minx:
            raise ValueError("MaxX should be greater than MinX.")

        if maxy <= miny:
            raise ValueError("MaxY should be greater than MinY.")

        self.minx = minx
        self.miny = miny
        self.maxx = maxx
        self.maxy = maxy

        self.element = etree.Element('{http://www.opengis.net/gml}Envelope')
        self.element.set('srsDimension', '2')
        self.element.set(
            'srsName',
            'http://www.opengis.net/gml/srs/epsg.xml#{:d}'.format(epsg))

        lower_corner = etree.Element('{http://www.opengis.net/gml}lowerCorner')
        lower_corner.text = '{:.06f} {:.06f}'.format(self.minx, self.miny)
        self.element.append(lower_corner)

        upper_corner = etree.Element('{http://www.opengis.net/gml}upperCorner')
        upper_corner.text = '{:.06f} {:.06f}'.format(self.maxx, self.maxy)
        self.element.append(upper_corner)

    def get_element(self):
        return self.element


class Point(AbstractLocation):
    """Class representing a point location."""

    def __init__(self, x, y, epsg=31370):
        """Initialise a Point.

        To initialise a Point using GPS coordinates in decimal degrees,
        use epsg=4326 and enter the longitude as `x` and the latitude as `y`.

        Parameters
        ----------
        x : float
            X coordinate (or longitude) of the point.
        y : float
            Y coordinate (or latitude) of the point.
        epsg : int, optional
            EPSG code of the coordinate reference system (CRS) of the
            coordinates specified in `x`, `y`. Defaults to
            31370 (Belgian Lambert72).

        """
        self.x = x
        self.y = y

        self.element = etree.Element('{http://www.opengis.net/gml}Point')
        self.element.set('srsDimension', '2')
        self.element.set(
            'srsName',
            'http://www.opengis.net/gml/srs/epsg.xml#{:d}'.format(epsg))

        coordinates = etree.Element('{http://www.opengis.net/gml}pos')
        coordinates.text = '{:.06f} {:.06f}'.format(self.x, self.y)
        self.element.append(coordinates)

    def get_element(self):
        return self.element


class GmlObject(AbstractLocation):
    """Class representing a raw GML location, f.ex. gml:Surface or
    gml:MultiSurface."""

    def __init__(self, gml_element):
        """Initialise a GmlObject.

        Initialise a GmlObject from an existing XML element representing a
        GML location.

        Parameters
        ----------
        gml_element : etree.Element or str or bytes
            XML element of the GML location, either as etree.Element, bytes or
            string representation.

        """
        if isinstance(gml_element, str):
            self.element = etree.fromstring(gml_element.encode('utf8'))
        elif isinstance(gml_element, bytes):
            self.element = etree.fromstring(gml_element)
        else:
            self.element = gml_element

    def get_element(self):
        return self.element


class Equals(AbstractBinarySpatialFilter):
    """Class representing a spatial Equals filter.

    A spatial Equals will return all points that are equal to another point.

    """

    def __init__(self, location):
        """Initialise an Equals filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the Equals
            filter.

        """
        super(Equals, self).__init__('Equals', location)


class Disjoint(AbstractBinarySpatialFilter):
    """Class representing a spatial Disjoint filter.

    A spatial Disjoint will return all points that are disjoint from a
    polygon or box location: i.e. that are not inside nor on the boundary.

    A spatial Disjoint is the inverse of a spatial Intersects.

    """

    def __init__(self, location):
        """Initialise a Disjoint filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the Disjoint
            filter.

        """
        super(Disjoint, self).__init__('Disjoint', location)


class Touches(AbstractBinarySpatialFilter):
    """Class representing a spatial Touches filter.

    A spatial Touches will return all points that touch a polygon or box
    location: i.e. that are on the boundary but not inside.

    """

    def __init__(self, location):
        """Initialise a Touches filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the Touches
            filter.

        """
        super(Touches, self).__init__('Touches', location)


class Within(AbstractBinarySpatialFilter):
    """Class representing a spatial Within filter.

    A spatial Within will return all points that are entirely within a
    polygon or box location (i.e. are not on the boundary).

    """

    def __init__(self, location):
        """Initialise a Within filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the Within
            filter.

        """
        super(Within, self).__init__('Within', location)


class Intersects(AbstractBinarySpatialFilter):
    """Class representing a spatial Intersects filter.

    A spatial Intersects will return all points that are within or on the
    boundary of a polygon or box location.

    A spatial Intersects is the inverse of a spatial Disjoint.

    """

    def __init__(self, location):
        """Initialise an Intersects filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the Intersects
            filter.

        """
        super(Intersects, self).__init__('Intersects', location)


class WithinDistance(AbstractLocationFilter):
    """Class representing a spatial DWithin filter.

    A spatial DWithin will return all points that are within a given
    distance of a certain location.

    """

    def __init__(self, location, distance, distance_unit='meter'):
        """Initialise a WithinDistance filter.

        Parameters
        ----------
        location : AbstractLocation
            An instance of a location to use as location for the
            WithinDistance filter.
        distance : float
            Amount of distance units to use for the filter.
        distance_unit : string, optional, defaults to 'meter'
            The distance unit of the value of `distance`.

        """
        self.location = location
        self.distance = distance
        self.distance_unit = distance_unit
        self.geom_column = ''

        self.element = etree.Element('{http://www.opengis.net/ogc}DWithin')

        geom = etree.Element('{http://www.opengis.net/ogc}PropertyName')
        geom.text = self.geom_column

        distance = etree.Element('{http://www.opengis.net/ogc}Distance')
        distance.set('units', self.distance_unit)
        distance.text = '{:.06f}'.format(self.distance)

        self.element.append(geom)
        self.element.append(location.get_element())
        self.element.append(distance)

    def set_geometry_column(self, geometry_column):
        self.geom_column = geometry_column
        geom = self.element.find('.//{http://www.opengis.net/ogc}PropertyName')
        geom.text = geometry_column

    def toXML(self):
        if self.geom_column == '':
            raise RuntimeError('Geometry column has not been set. Use '
                               '"set_geometry_column" to set it.')
        return self.element


class GmlFilter(AbstractLocationFilter):
    """Class for construction a spatial filter expression from a GML
    3.1.1 document.
    """

    def __init__(self, gml, location_filter, location_filter_kwargs=None,
                 combinator=Or):
        """Initialise a spatial filter expression from a GML 3.1.1 string.

        Parameters
        ----------
        gml : str
            Either a string representation of the GML document to parse,
            or a path to a GML file on disk.
        location_filter : class<AbstractLocationFilter>
            Location filter to use for the geometries in the GML document.
        location_filter_kwargs : dict, optional
            Keyword-based arguments to pass to the `location_filter` on
            initialisation (with the exception of the `location` parameter,
            which is automatically parsed from the GML). Can be skipped in
            cases where the location_filter takes no extra arguments besides
            location.
        combinator : class<BinaryLogicOpType>, optional, defaults to Or
            One of (Or, And) used to combine filters for different geometries
            in the GML document.

        Raises
        ------
        ValueError
            When no geometries could be parsed from the given GML record.

        """
        self.gml = gml
        self.subelements = OrderedDict()

        if location_filter_kwargs is None:
            location_filter_kwargs = {}

        self._parse_gml()

        if len(self.subelements) == 1:
            self.element = location_filter(
                GmlObject(list(self.subelements)[0]),
                **location_filter_kwargs)
        else:
            self.element = combinator(
                [location_filter(GmlObject(e), **location_filter_kwargs)
                 for e in self.subelements])

    def _dedup_multi(self, tree, xpath_single, xpath_multi):
        """Parse single and multi* geometries from the same type from the
        GML tree. Deduplicates in the sense that single geometries inside a
        multi* geometry will be discarded in favor of the latter.

        Parameters
        ----------
        tree : etree.ElementTree
            XML tree of the GML to parse.
        xpath_single : str
            XPath of the single geometries to parse.
        xpath_multi : str
            XPath of the multi* geometries to parse.

        Returns
        -------
        single, multi : set of etree.Element, set of etree.Element
            Sets of the parsed single and multi geometry Elements.

        """
        single = OrderedDict.fromkeys(tree.findall(xpath_single))
        multi = OrderedDict.fromkeys(tree.findall(xpath_multi))

        for m in multi:
            s_in_m = OrderedDict.fromkeys(m.findall(xpath_single))
            single = OrderedDict.fromkeys(e for e in single if e not in s_in_m)

        return single, multi

    def _parse_gml(self):
        """Checks the gml parameter and tries parsing the file.

        Raises
        ------
        ValueError
            When the file could not be parsed.

        """
        gml_tree = None
        try:
            if isinstance(self.gml, str):
                gml_tree = etree.fromstring(self.gml.encode('utf8'))
            elif isinstance(self.gml, bytes):
                gml_tree = etree.fromstring(self.gml)
        except etree.ParseError as error:
            if os.path.isfile(self.gml):
                with open(self.gml, 'r') as gml_file:
                    gml_tree = etree.fromstring(gml_file.read().encode('utf8'))
            else:
                raise error
        finally:
            if gml_tree is not None:
                self._parse_gml_tree(gml_tree)
            else:
                raise ValueError('Failed to parse GML file.')

    def _parse_gml_tree(self, gml_tree):
        """Parse the GML tree and add subelements for geometries.

        Parameters
        ----------
        gml_tree : etree.ElementTree
            XML tree of the GML to parse.

        Raises
        ------
        ValueError
            When no geometries could be parsed from the given GML record.

        """
        points, multipoints = self._dedup_multi(
            gml_tree,
            './/{http://www.opengis.net/gml}Point',
            './/{http://www.opengis.net/gml}MultiPoint'
        )

        self.subelements.update(points)
        self.subelements.update(multipoints)

        linestrings, multicurves = self._dedup_multi(
            gml_tree,
            './/{http://www.opengis.net/gml}LineString',
            './/{http://www.opengis.net/gml}MultiCurve'
        )

        self.subelements.update(linestrings)
        self.subelements.update(multicurves)

        polygons, multisurfaces = self._dedup_multi(
            gml_tree,
            './/{http://www.opengis.net/gml}Polygon',
            './/{http://www.opengis.net/gml}MultiSurface')

        self.subelements.update(polygons)
        self.subelements.update(multisurfaces)

        if len(self.subelements) == 0:
            raise ValueError('Failed to extract geometries from GML file.')

    def set_geometry_column(self, geometry_column):
        if len(self.subelements) == 1:
            self.element.set_geometry_column(geometry_column)
        else:
            for sub_element in self.element.operations:
                sub_element.set_geometry_column(geometry_column)

    def toXML(self):
        return self.element.toXML()


class GeometryFilter(GmlFilter):
    """Class for construction a spatial filter expression from any Geometry
    Fiona can convert to a GML 3.1.1 document.
    """

    def __init__(self, geometry, location_filter,
                 location_filter_kwargs=None, combinator=Or):
        """Initialise a spatial filter expression from a given Geometry.

        Parameters
        ----------
        geometry : str
            A path to a vector file on disk.
        location_filter : class<AbstractLocationFilter>
            Location filter to use for the geometries in the GML document.
        location_filter_kwargs : dict, optional
            Keyword-based arguments to pass to the `location_filter` on
            initialisation (with the exception of the `location` parameter,
            which is automatically parsed from the geomery). Can be skipped in
            cases where the location_filter takes no extra arguments besides
            location.
        combinator : class<BinaryLogicOpType>, optional, defaults to Or
            One of (Or, And) used to combine filters for different geometries
            in the geometry document.

        Raises
        ------
        ValueError
            When no geometries could be parsed from the given geometry.
        """
        gml = self._parse_geometry(geometry)

        super().__init__(gml, location_filter, location_filter_kwargs,
                         combinator)

    @staticmethod
    def _parse_geometry(geometry):
        """Checks the geometry and parse the file if possible.

        Raises
        ------
        ValueError
            When the file could not be parsed.
        """
        try:
            import fiona

            # Check if the GML driver and write mode are supported and check
            # if version of Fiona supports GML 3.1.1 output (Fiona >= 1.8.18)
            drivers = fiona.supported_drivers
            if 'w' not in drivers.get('GML', '') or \
                    LooseVersion(fiona.__version__) < LooseVersion('1.8.18'):
                raise UserWarning('Please use module Fiona version 1.8.18 '
                                  'or higher')
        except ImportError:
            raise ImportError('No module named fiona. GeometryFilter '
                              'requires fiona to be installed.')
        else:
            # Open the geometry as a collection containing multiple records
            with fiona.open(geometry, 'r') as c:
                gml_blob = BytesIO()
                # Assume the geometry is loaded from file
                layer = os.path.basename(os.path.splitext(geometry)[0])
                with fiona.open(gml_blob, 'w', layer=layer, driver='GML',
                                schema=c.schema, crs=c.crs,
                                FORMAT='GML3') as gml:
                    for r in c:
                        gml.write(r)
                gml_blob.seek(0)
                # Set the gml attribute used by the parent class
                gml = gml_blob.read()
                return gml


class GeopandasFilter(GmlFilter):
    """Class for construction a spatial filter expression from a
    GeoPandas object.
    """

    def __init__(self, geodataframe, location_filter,
                 location_filter_kwargs=None, combinator=Or):
        """Initialise a spatial filter expression from a GeoPandas
        GeoDataFrame.

        Parameters
        ----------
        geodataframe : GeoPandas.GeoDataFrame
            A GeoDataFrame with a valid crs.
        location_filter : class<AbstractLocationFilter>
            Location filter to use for the geometries in the GML document.
        location_filter_kwargs : dict, optional
            Keyword-based arguments to pass to the `location_filter` on
            initialisation (with the exception of the `location` parameter,
            which is automatically parsed from the geomery). Can be skipped in
            cases where the location_filter takes no extra arguments besides
            location.
        combinator : class<BinaryLogicOpType>, optional, defaults to Or
            One of (Or, And) used to combine filters for different geometries
            in the geometry document.

        Raises
        ------
        ValueError
            When no geometries could be parsed from the given geometry.
        """
        gml = self._parse_geometry(geodataframe)

        super().__init__(gml, location_filter, location_filter_kwargs,
                         combinator)

    @staticmethod
    def _parse_geometry(gdf):
        """Checks the geometry and parse the file if possible.

        Raises
        ------
        ValueError
            When the file could not be parsed.
        """
        try:
            import geopandas
        except ImportError:
            raise ImportError('No module named GeoPandas. GeopandasFilter '
                              'requires geopandas to be installed.')
        else:
            if not isinstance(gdf, geopandas.GeoDataFrame):
                raise TypeError("pydov only supports geopandas.GeoDataFrame "
                                "to define a spatial query.")

            if not gdf.crs:
                raise AttributeError("geopandas.GeoDataFrame is missing a "
                                     "CRS definition")

            with BytesIO() as gml_blob:
                gdf.to_file(gml_blob, driver="GML", FORMAT="GML3",
                            layer="geodataframe")
                gml_blob.seek(0)
                gml = gml_blob.read()
            return gml



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import find_packages, setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()
with open('requirements_dev.txt') as f:
    # ignore the general requirements
    requirements_dev = f.read().splitlines()[1:]
with open('requirements_doc.txt') as f:
    requirements_doc = f.read().splitlines()
with open('requirements_vectorfile.txt') as f:
    requirements_vectorfile = f.read().splitlines()

setup(
    name='pydov',
    version='2.1.0',
    description=("A Python package to download data from Databank Ondergrond "
                 "Vlaanderen (DOV)."),
    long_description=readme,
    long_description_content_type='text/markdown',
    author="DOV-Vlaanderen",
    author_email='dov@vlaanderen.be',
    url='https://github.com/DOV-Vlaanderen/pydov',
    packages=find_packages(
        include=['pydov']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='pydov',
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Natural Language :: Dutch',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Scientific/Engineering',
    ],
    test_suite='tests',
    tests_require=requirements_dev,
    extras_require={
        'docs': requirements_doc,
        'devs': requirements_dev,
        'vectorfile': requirements_vectorfile
    }
)



"""Module grouping common tests for all search modules."""

import numpy as np
import pytest

from pydov.search.bodemlocatie import BodemlocatieSearch
from pydov.search.bodemmonster import BodemmonsterSearch
from pydov.search.bodemobservatie import BodemobservatieSearch
from pydov.search.bodemsite import BodemsiteSearch
from pydov.search.boring import BoringSearch
from pydov.search.grondmonster import GrondmonsterSearch
from pydov.search.grondwaterfilter import GrondwaterFilterSearch
from pydov.search.grondwatermonster import GrondwaterMonsterSearch
from pydov.search.grondwatervergunning import GrondwaterVergunningSearch
from pydov.search.interpretaties import (
    FormeleStratigrafieSearch, GecodeerdeLithologieSearch,
    GeotechnischeCoderingSearch, HydrogeologischeStratigrafieSearch,
    InformeleHydrogeologischeStratigrafieSearch, InformeleStratigrafieSearch,
    LithologischeBeschrijvingenSearch, QuartairStratigrafieSearch)
from pydov.search.sondering import SonderingSearch
from pydov.util.errors import InvalidSearchParameterError
from pydov.util.location import Point, WithinDistance
from tests.abstract import ServiceCheck

search_objects = [BodemsiteSearch(),
                  BodemlocatieSearch(),
                  BodemobservatieSearch(),
                  BodemmonsterSearch(),
                  BoringSearch(),
                  SonderingSearch(),
                  GrondwaterFilterSearch(),
                  GrondwaterMonsterSearch(),
                  GrondwaterVergunningSearch(),
                  FormeleStratigrafieSearch(),
                  InformeleHydrogeologischeStratigrafieSearch(),
                  GeotechnischeCoderingSearch(),
                  QuartairStratigrafieSearch(),
                  InformeleStratigrafieSearch(),
                  HydrogeologischeStratigrafieSearch(),
                  GecodeerdeLithologieSearch(),
                  LithologischeBeschrijvingenSearch(),
                  GrondmonsterSearch()]


@pytest.mark.parametrize("objectsearch", search_objects)
def test_get_description(mp_wfs, objectsearch):
    """Test the get_description method.

    Test whether the method returns a non-empty string.

    Parameters
    ----------
    mp_wfs : pytest.fixture
        Monkeypatch the call to the remote GetCapabilities request.
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    description = objectsearch.get_description()

    assert isinstance(description, str)
    assert len(description) > 0


@pytest.mark.online
@pytest.mark.skipif(not ServiceCheck.service_ok(),
                    reason="DOV service is unreachable")
@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_location(objectsearch):
    """Test the get_description method.

    Test whether the method returns a non-empty string.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    objectsearch.search(location=WithinDistance(Point(100000, 100000), 100))


@pytest.mark.online
@pytest.mark.skipif(not ServiceCheck.service_ok(),
                    reason="DOV service is unreachable")
@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_maxfeatures(objectsearch):
    """Test the search method with a max_features parameter.

    Test whether no error is raised.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    objectsearch.search(location=WithinDistance(Point(100000, 100000), 100),
                        max_features=10)


@pytest.mark.online
@pytest.mark.skipif(not ServiceCheck.service_ok(),
                    reason="DOV service is unreachable")
@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_maxfeatures_only(objectsearch):
    """Test the search method with only the max_features parameter.

    Test whether no error is raised.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    objectsearch.search(max_features=1)


@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_nolocation_noquery(objectsearch):
    """Test the search method without providing a location or a query.

    Test whether an InvalidSearchParameterError is raised.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    with pytest.raises(InvalidSearchParameterError):
        objectsearch.search(location=None, query=None)


@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_both_location_query_wrongquerytype(objectsearch):
    """Test the search method providing both a location and a query,
    using a query with an invalid type.

    Test whether an InvalidSearchParameterError is raised.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    with pytest.raises(InvalidSearchParameterError):
        objectsearch.search(location=(1, 2, 3, 4),
                            query='computer says no')


@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_query_wrongtype(objectsearch):
    """Test the search method with the query parameter using a wrong
    query type.

    Test whether an InvalidSearchParameterError is raised.

    Parameters
    ----------
    objectsearch : pytest.fixture
        An instance of a subclass of AbstractTestSearch to perform search
        operations on the corresponding DOV type.

    """
    with pytest.raises(InvalidSearchParameterError):
        objectsearch.search(query='computer says no')


@pytest.mark.online
@pytest.mark.skipif(not ServiceCheck.service_ok(),
                    reason="DOV service is unreachable")
@pytest.mark.parametrize("objectsearch", search_objects)
def test_search_wfs_no_primary_key(objectsearch):
    """Test the search method with return fields from WFS but not
    including the primary key column.

    Test whether the output dataframe contains the resolved WFS data.
    """
    wfs_return_fields = [
        f['name'] for f in objectsearch.get_fields().values()
        if f['cost'] == 1 and f['type'] == 'float' and f['notnull']
        and 'pkey' not in f['name']]

    if len(wfs_return_fields) > 0:
        field = wfs_return_fields[0]

        df = objectsearch.search(max_features=1, return_fields=(field,))
        assert not np.isnan(df[field][0])



# -*- coding: utf-8 -*-
"""Module containing the DOV data type for bodemlocaties, including
subtypes."""
from pydov.types.fields import WfsField, XmlField

from .abstract import AbstractDovType


class Bodemlocatie(AbstractDovType):
    """Class representing the DOV data type for bodemlocaties."""

    subtypes = []

    fields = [
        WfsField(name='pkey_bodemlocatie', source_field='Bodemlocatiefiche',
                 datatype='string'),
        WfsField(name='pkey_bodemsite', source_field='Bodemsitefiche',
                 datatype='string'),
        WfsField(name='naam', source_field='Naam', datatype='string'),
        WfsField(name='type', source_field='Type', datatype='string'),
        WfsField(name='waarnemingsdatum', source_field='Datum',
                 datatype='date'),
        WfsField(name='doel', source_field='Doel', datatype='string'),
        WfsField(name='x', source_field='X_mL72', datatype='float'),
        WfsField(name='y', source_field='Y_mL72', datatype='float'),
        WfsField(name='mv_mtaw', source_field='mv_mTAW', datatype='float'),
        WfsField(name='erfgoed', source_field='Erfgoed', datatype='boolean'),
        WfsField(name='bodemstreek', source_field='Bodemstreek',
                 datatype='string'),
        XmlField(name='invoerdatum',
                 source_xpath='/bodemlocatie/invoerdatum',
                 definition='Datum van invoer van de bodemlocatie.',
                 datatype='date'),
        XmlField(name='educatieve_waarde',
                 source_xpath='/bodemlocatie/educatieve_waarde',
                 definition='Educatieve waarde van de bodemlocatie.',
                 datatype='string')
    ]

    pkey_fieldname = 'Bodemlocatiefiche'

    def __init__(self, pkey):
        """Initialisation.

        Parameters
        ----------
        pkey : str
            Permanent key of the Bodemlocatie, being a URI of the form
            `https://www.dov.vlaanderen.be/data/bodemlocatie/<id>`.

        """
        super().__init__('bodemlocatie', pkey)



# -*- coding: utf-8 -*-
"""Module containing the search classes to retrieve DOV groundwater screen
 data."""
import pandas as pd
from owslib.fes import And, Not, PropertyIsNull

from ..types.grondwaterfilter import GrondwaterFilter
from .abstract import AbstractSearch


class GrondwaterFilterSearch(AbstractSearch):
    """Search class to retrieve information about groundwater screens
    (GrondwaterFilter).
    """

    def __init__(self, objecttype=GrondwaterFilter):
        """Initialisation.

        Parameters
        ----------
        objecttype : subclass of pydov.types.abstract.AbstractDovType
            Reference to a class representing the GrondwaterFilter type.
            Optional: defaults to the GrondwaterFilter type containing the
            fields described in the documentation.

        """
        super(GrondwaterFilterSearch,
              self).__init__('gw_meetnetten:meetnetten', objecttype)

    def search(self, location=None, query=None, sort_by=None,
               return_fields=None, max_features=None):
        """Search for objects of this type. Provide `location` and/or
        `query` and/or `max_features`.
        When `return_fields` is None, all fields are returned.

        Excludes 'empty' filters (i.e. Putten without Filters) by extending
        the `query` with a not-null check on pkey_filter.

        Parameters
        ----------
        location : pydov.util.location.AbstractLocationFilter or \
                   owslib.fes.BinaryLogicOpType<AbstractLocationFilter> or \
                   owslib.fes.UnaryLogicOpType<AbstractLocationFilter>
            Location filter limiting the features to retrieve. Can either be a
            single instance of a subclass of AbstractLocationFilter, or a
            combination using And, Or, Not of AbstractLocationFilters.
        query : owslib.fes.OgcExpression
            OGC filter expression to use for searching. This can contain any
            combination of filter elements defined in owslib.fes. The query
            should use the fields provided in `get_fields()`. Note that not
            all fields are currently supported as a search parameter.
        sort_by : owslib.fes.SortBy, optional
            List of properties to sort by.
        return_fields : list<str> or tuple<str> or set<str>
            A list of fields to be returned in the output data. This should
            be a subset of the fields provided in `get_fields()`. Note that
            not all fields are currently supported as return fields.
        max_features : int
            Limit the maximum number of features to request.

        Returns
        -------
        pandas.core.frame.DataFrame
            DataFrame containing the output of the search query.

        Raises
        ------
        pydov.util.errors.InvalidSearchParameterError
            When not one of `location` or `query` or `max_features` is
            provided.

        pydov.util.errors.InvalidFieldError
            When at least one of the fields in `return_fields` is unknown.

            When a field that is only accessible as return field is used as
            a query parameter.

            When a field that can only be used as a query parameter is used as
            a return field.

        pydov.util.errors.FeatureOverflowError
            When the number of features to be returned is equal to the
            maxFeatures limit of the WFS server.

        AttributeError
            When the argument supplied as return_fields is not a list,
            tuple or set.

        NotImplementedError
            This is an abstract method that should be implemented in a
            subclass.

        """
        self._pre_search_validation(location, query, sort_by, return_fields,
                                    max_features)

        exclude_empty_filters = Not([PropertyIsNull(
                                     propertyname='pkey_filter')])

        if query is not None:
            query = And([query, exclude_empty_filters])
        else:
            query = exclude_empty_filters

        fts = self._search(location=location, query=query, sort_by=sort_by,
                           return_fields=return_fields,
                           max_features=max_features)

        gw_filters = self._type.from_wfs(fts, self._wfs_namespace)

        df = pd.DataFrame(
            data=self._type.to_df_array(gw_filters, return_fields),
            columns=self._type.get_field_names(return_fields))

        return df



# -*- coding: utf-8 -*-
"""Module grouping all classes related to pydov field definitions."""


class XsdType(object):
    """Class for specifying an XSD type from an XSD schema. This will be
    resolved at runtime in a list of possible values and their definitions."""

    def __init__(self, xsd_schema, typename):
        """Initialise a XSD type reference.

        Parameters
        ----------
        xsd_schema : str
            URL of XSD schema record containing the specified typename.
        typename : str
            Name of the type.

        """
        self.xsd_schema = xsd_schema
        self.typename = typename


class AbstractField(dict):
    """Abstract base class for pydov field definitions. Not to be
    instantiated directly."""

    def __init__(self, name, source, datatype, **kwargs):
        """Initialise a field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source : one of 'wfs', 'xml', 'custom'
            Source of this field.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(AbstractField, self).__init__(**kwargs)
        self.__setitem__('name', name)
        self.__setitem__('source', source)
        self.__setitem__('type', datatype)


class WfsField(AbstractField):
    """Class for a field available in the WFS service."""

    def __init__(self, name, source_field, datatype):
        """Initialise a WFS field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source_field : str
            Name of this field in the source WFS service.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(WfsField, self).__init__(name, 'wfs', datatype)
        self.__setitem__('sourcefield', source_field)


class _WfsInjectedField(WfsField):
    """Class for a field available in the WFS service, but not included in the
    default dataframe output."""

    def __init__(self, name, datatype):
        """Initialise a WFS injected field.

        This is a field not normally present in the dataframe, but usable as
        a query and returnfield as it is available in the WFS service.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.

        """
        super(_WfsInjectedField, self).__init__(name, name, datatype)
        self.__setitem__('wfs_injected', True)


class XmlField(AbstractField):
    """Class for a field available in the XML document."""

    def __init__(self, name, source_xpath, datatype, definition='',
                 notnull=False, xsd_type=None):
        """Initialise an XML field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        source_xpath : str
            XPath expression of the values of this field in the source XML
            document.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.
        definition : str, optional
            Definition of this field.
        notnull : bool, optional, defaults to False
            True if this field is always present (mandatory), False otherwise.
        xsd_type : pydov.types.abstract.XsdType, optional
            XSD type associated with this field.

        """
        super(XmlField, self).__init__(name, 'xml', datatype)

        self.__setitem__('sourcefield', source_xpath)
        self.__setitem__('definition', definition)
        self.__setitem__('notnull', notnull)

        if xsd_type is not None:
            self.__setitem__('xsd_schema', xsd_type.xsd_schema)
            self.__setitem__('xsd_type', xsd_type.typename)


class _CustomField(AbstractField):
    """Class for a custom field, created explicitly in pydov."""

    def __init__(self, name, datatype, definition='', notnull=False):
        """Initialise a custom field.

        Parameters
        ----------
        name : str
            Name of this field in the return dataframe.
        datatype : one of 'string', 'integer', 'float', 'date', 'datetime' \
                   or 'boolean'
            Datatype of the values of this field in the return dataframe.
        definition : str, optional
            Definition of this field.
        notnull : bool, optional, defaults to False
            True if this field is always present (mandatory), False otherwise.

        """
        super(_CustomField, self).__init__(name, 'custom', datatype)
        self.__setitem__('definition', definition)
        self.__setitem__('notnull', notnull)
